import json

from six import StringIO

import pytest

from cwltool.main import main
from cwltool import load_tool
from .util import get_data, needs_docker

override_parameters = [
    ([get_data('tests/override/echo.cwl'),
      get_data('tests/override/echo-job.yml')],
     {"out": "zing hello1\n"}
     ),
    (["--overrides",
      get_data('tests/override/ov.yml'),
      get_data('tests/override/echo.cwl'),
      get_data('tests/override/echo-job.yml')],
     {"out": "zing hello2\n"}
     ),
    ([get_data('tests/override/echo.cwl'),
      get_data('tests/override/echo-job-ov.yml')],
     {"out": "zing hello3\n"}
     ),
    ([get_data('tests/override/echo-job-ov2.yml')],
     {"out": "zing hello4\n"}
     ),
    (["--overrides",
      get_data('tests/override/ov.yml'),
      get_data('tests/override/echo-wf.cwl'),
      get_data('tests/override/echo-job.yml')],
     {"out": "zing hello2\n"}
     ),
    (["--overrides",
      get_data('tests/override/ov2.yml'),
      get_data('tests/override/echo-wf.cwl'),
      get_data('tests/override/echo-job.yml')],
     {"out": "zing hello5\n"}
     ),
    (["--overrides",
      get_data('tests/override/ov3.yml'),
      get_data('tests/override/echo-wf.cwl'),
      get_data('tests/override/echo-job.yml')],
     {"out": "zing hello6\n"}
     ),
    ([get_data('tests/override/env-tool_v1.1.cwl'),
      get_data('tests/override/env-tool_cwl-requirement_override.yaml')],
     {"value": "hello test env"}
     ),
    (["--enable-dev",
      get_data('tests/override/env-tool_cwl-requirement_override_default.yaml')],
     {"value": "hello test env"}
     ),
]

@needs_docker
@pytest.mark.parametrize('parameters,result', override_parameters)
def test_overrides(parameters, result):
    sio = StringIO()

    assert main(parameters, stdout=sio) == 0
    assert json.loads(sio.getvalue()) == result


failing_override_parameters = [
    ([get_data('tests/override/env-tool.cwl'),
      get_data('tests/override/env-tool_cwl-requirement_override.yaml')],
     "`cwl:requirements` in the input object is not part of CWL v1.0. You can "
     "adjust to use `cwltool:overrides` instead; or you can set the cwlVersion to "
     "v1.1"
     ),
    ([get_data('tests/override/env-tool_v1.1.0-dev1.cwl'),
      get_data('tests/override/env-tool_cwl-requirement_override.yaml')],
     "Version 'v1.1.0-dev1' is a development or deprecated version.\n"
     " Update your document to a stable version (v1.0, v1.1) or use --enable-dev to "
     "enable support for development and deprecated versions."
     ),
    ([get_data('tests/override/env-tool_cwl-requirement_override_default_wrongver.yaml')],
     "`cwl:requirements` in the input object is not part of CWL v1.0. You can "
     "adjust to use `cwltool:overrides` instead; or you can set the cwlVersion to "
     "v1.1 or greater."
     ),
    ([get_data('tests/override/env-tool_cwl-requirement_override_default.yaml')],
     "Version 'v1.1.0-dev1' is a development or deprecated version.\n"
     " Update your document to a stable version (v1.0, v1.1) or use --enable-dev to "
     "enable support for development and deprecated versions."
     ),
]

@needs_docker
@pytest.mark.parametrize('parameters,expected_error', failing_override_parameters)
def test_overrides_fails(parameters, expected_error):
    sio = StringIO()

    assert main(parameters, stderr=sio) == 1
    stderr = sio.getvalue()
    assert expected_error in stderr, stderr



from __future__ import absolute_import

from collections import namedtuple
from typing import Any, Dict

from typing_extensions import Text  # pylint: disable=unused-import
# move to a regular typing import when Python 3.3-3.6 is no longer supported

from .errors import WorkflowException


MutationState = namedtuple("MutationTracker", ["generation", "readers", "stepname"])

_generation = "http://commonwl.org/cwltool#generation"

class MutationManager(object):
    """Lock manager for checking correctness of in-place update of files.

    Used to validate that in-place file updates happen sequentially, and that a
    file which is registered for in-place update cannot be read or updated by
    any other steps.

    """

    def __init__(self):  # type: () -> None
        """Initialize."""
        self.generations = {}  # type: Dict[Text, MutationState]

    def register_reader(self, stepname, obj):
        # type: (Text, Dict[Text, Any]) -> None
        loc = obj["location"]
        current = self.generations.get(loc, MutationState(0, [], ""))
        obj_generation = obj.get(_generation, 0)

        if obj_generation != current.generation:
            raise WorkflowException(
                "[job {}] wants to read {} from generation {} but current "
                "generation is {}(last updated by {})".format(
                    stepname, loc, obj_generation, current.generation, current.stepname))

        current.readers.append(stepname)
        self.generations[loc] = current

    def release_reader(self, stepname, obj):
        # type: (Text, Dict[Text, Any]) -> None
        loc = obj["location"]
        current = self.generations.get(loc, MutationState(0, [], ""))
        obj_generation = obj.get(_generation, 0)

        if obj_generation != current.generation:
            raise WorkflowException(
                "[job {}] wants to release reader on {} from generation {}"
                " but current generation is {} (last updated by {})".format(
                    stepname, loc, obj_generation, current.generation,
                    current.stepname))

        self.generations[loc].readers.remove(stepname)

    def register_mutation(self, stepname, obj):
        # type: (Text, Dict[Text, Any]) -> None
        loc = obj["location"]
        current = self.generations.get(loc, MutationState(0, [], ""))
        obj_generation = obj.get(_generation, 0)

        if len(current.readers) > 0:
            raise WorkflowException(
                "[job {}] wants to modify {} but has readers: {}".format(
                    stepname, loc, current.readers))

        if obj_generation != current.generation:
            raise WorkflowException(
                "[job {}] wants to modify {} from generation {} but current "
                "generation is {} (last updated by {})".format(
                    stepname, loc, obj_generation, current.generation, current.stepname))

        self.generations[loc] = MutationState(current.generation+1, current.readers, stepname)

    def set_generation(self, obj):  # type: (Dict[Text, Text]) -> None
        loc = obj["location"]
        current = self.generations.get(loc, MutationState(0, [], ""))
        obj[_generation] = current.generation

    def unset_generation(self, obj):  # type: (Dict[Text, Text]) -> None
        obj.pop(_generation, None)



from cwltool import command_line_tool
from cwltool.utils import windows_default_container_id
from cwltool.context import RuntimeContext


# Test to check warning when default docker Container is used on Windows
def test_default_docker_warning(mocker):
    mocker.patch("cwltool.command_line_tool.onWindows", return_value=True)
    mocker.patch("cwltool.command_line_tool._logger")

    class TestCommandLineTool(command_line_tool.CommandLineTool):
        def __init__(self, **kwargs):
            self.requirements = []
            self.hints = []

        def find_default_container(self, args, builder):
            return windows_default_container_id

    tool = TestCommandLineTool()
    tool.make_job_runner(RuntimeContext({
        "find_default_container": lambda x: "frolvlad/alpine-bash"}))

    command_line_tool._logger.warning.assert_called_with(
        command_line_tool.DEFAULT_CONTAINER_MSG,
        windows_default_container_id,
        windows_default_container_id)



"""Static checking of CWL workflow connectivity."""
from collections import namedtuple
from typing import Any, Dict, List, MutableMapping, MutableSequence, Optional

import six
from schema_salad import validate
from schema_salad.sourceline import SourceLine, bullets, strip_dup_lineno
from typing_extensions import Text  # pylint: disable=unused-import
# move to a regular typing import when Python 3.3-3.6 is no longer supported

from .errors import WorkflowException
from .loghandler import _logger
from .process import shortname
from .utils import json_dumps


def _get_type(tp):
    # type: (Any) -> Any
    if isinstance(tp, MutableMapping):
        if tp.get("type") not in ("array", "record", "enum"):
            return tp["type"]
    return tp

def check_types(srctype, sinktype, linkMerge, valueFrom):
    # type: (Any, Any, Optional[Text], Optional[Text]) -> Text
    """
    Check if the source and sink types are correct.

    Acceptable types are "pass", "warning", or "exception".
    """
    if valueFrom is not None:
        return "pass"
    if linkMerge is None:
        if can_assign_src_to_sink(srctype, sinktype, strict=True):
            return "pass"
        if can_assign_src_to_sink(srctype, sinktype, strict=False):
            return "warning"
        return "exception"
    if linkMerge == "merge_nested":
        return check_types({"items": _get_type(srctype), "type": "array"},
                           _get_type(sinktype), None, None)
    if linkMerge == "merge_flattened":
        return check_types(merge_flatten_type(_get_type(srctype)), _get_type(sinktype), None, None)
    raise WorkflowException(u"Unrecognized linkMerge enum '{}'".format(linkMerge))


def merge_flatten_type(src):
    # type: (Any) -> Any
    """Return the merge flattened type of the source type."""
    if isinstance(src, MutableSequence):
        return [merge_flatten_type(t) for t in src]
    if isinstance(src, MutableMapping) and src.get("type") == "array":
        return src
    return {"items": src, "type": "array"}


def can_assign_src_to_sink(src, sink, strict=False):  # type: (Any, Any, bool) -> bool
    """
    Check for identical type specifications, ignoring extra keys like inputBinding.

    src: admissible source types
    sink: admissible sink types

    In non-strict comparison, at least one source type must match one sink type.
    In strict comparison, all source types must match at least one sink type.
    """
    if src == "Any" or sink == "Any":
        return True
    if isinstance(src, MutableMapping) and isinstance(sink, MutableMapping):
        if sink.get("not_connected") and strict:
            return False
        if src["type"] == "array" and sink["type"] == "array":
            return can_assign_src_to_sink(src["items"], sink["items"], strict)
        if src["type"] == "record" and sink["type"] == "record":
            return _compare_records(src, sink, strict)
        if src["type"] == "File" and sink["type"] == "File":
            for sinksf in sink.get("secondaryFiles", []):
                if not [1 for srcsf in src.get("secondaryFiles", []) if sinksf == srcsf]:
                    if strict:
                        return False
            return True
        return can_assign_src_to_sink(src["type"], sink["type"], strict)
    if isinstance(src, MutableSequence):
        if strict:
            for this_src in src:
                if not can_assign_src_to_sink(this_src, sink):
                    return False
            return True
        for this_src in src:
            if can_assign_src_to_sink(this_src, sink):
                return True
        return False
    if isinstance(sink, MutableSequence):
        for this_sink in sink:
            if can_assign_src_to_sink(src, this_sink):
                return True
        return False
    return bool(src == sink)


def _compare_records(src, sink, strict=False):
    # type: (MutableMapping[Text, Any], MutableMapping[Text, Any], bool) -> bool
    """
    Compare two records, ensuring they have compatible fields.

    This handles normalizing record names, which will be relative to workflow
    step, so that they can be compared.
    """
    def _rec_fields(rec):  # type: (MutableMapping[Text, Any]) -> MutableMapping[Text, Any]
        out = {}
        for field in rec["fields"]:
            name = shortname(field["name"])
            out[name] = field["type"]
        return out

    srcfields = _rec_fields(src)
    sinkfields = _rec_fields(sink)
    for key in six.iterkeys(sinkfields):
        if (not can_assign_src_to_sink(
                srcfields.get(key, "null"), sinkfields.get(key, "null"), strict)
                and sinkfields.get(key) is not None):
            _logger.info("Record comparison failure for %s and %s\n"
                         "Did not match fields for %s: %s and %s",
                         src["name"], sink["name"], key, srcfields.get(key),
                         sinkfields.get(key))
            return False
    return True

def missing_subset(fullset, subset): # type: (List[Any], List[Any]) -> List[Any]
    missing = []
    for i in subset:
        if i not in fullset:
            missing.append(i)
    return missing

def static_checker(workflow_inputs, workflow_outputs, step_inputs, step_outputs, param_to_step):
    # type: (List[Dict[Text, Any]], List[Dict[Text, Any]], List[Dict[Text, Any]], List[Dict[Text, Any]], Dict[Text, Dict[Text, Any]]) -> None
    """Check if all source and sink types of a workflow are compatible before run time."""
    # source parameters: workflow_inputs and step_outputs
    # sink parameters: step_inputs and workflow_outputs

    # make a dictionary of source parameters, indexed by the "id" field
    src_parms = workflow_inputs + step_outputs
    src_dict = {}
    for parm in src_parms:
        src_dict[parm["id"]] = parm

    step_inputs_val = check_all_types(src_dict, step_inputs, "source")
    workflow_outputs_val = check_all_types(src_dict, workflow_outputs, "outputSource")

    warnings = step_inputs_val["warning"] + workflow_outputs_val["warning"]
    exceptions = step_inputs_val["exception"] + workflow_outputs_val["exception"]

    warning_msgs = []
    exception_msgs = []
    for warning in warnings:
        src = warning.src
        sink = warning.sink
        linkMerge = warning.linkMerge
        sinksf = sorted([p["pattern"] for p in sink.get("secondaryFiles", []) if p.get("required", True)])
        srcsf = sorted([p["pattern"] for p in src.get("secondaryFiles", [])])
        # Every secondaryFile required by the sink, should be declared
        # by the source
        missing = missing_subset(srcsf, sinksf)
        if missing:
            msg1 = "Parameter '%s' requires secondaryFiles %s but" % (shortname(sink["id"]), missing)
            msg3 = SourceLine(src, "id").makeError(
                "source '%s' does not provide those secondaryFiles." % (shortname(src["id"])))
            msg4 = SourceLine(src.get("_tool_entry", src), "secondaryFiles").makeError("To resolve, add missing secondaryFiles patterns to definition of '%s' or" % (shortname(src["id"])))
            msg5 = SourceLine(sink.get("_tool_entry", sink), "secondaryFiles").makeError("mark missing secondaryFiles in definition of '%s' as optional." % shortname(sink["id"]))
            msg = SourceLine(sink).makeError("%s\n%s" % (msg1, bullets([msg3, msg4, msg5], "  ")))
        elif sink.get("not_connected"):
            msg = SourceLine(sink, "type").makeError(
                "'%s' is not an input parameter of %s, expected %s"
                % (shortname(sink["id"]), param_to_step[sink["id"]]["run"],
                   ", ".join(shortname(s["id"])
                             for s in param_to_step[sink["id"]]["inputs"]
                             if not s.get("not_connected"))))
        else:
            msg = SourceLine(src, "type").makeError(
                "Source '%s' of type %s may be incompatible"
                % (shortname(src["id"]), json_dumps(src["type"]))) + "\n" + \
                SourceLine(sink, "type").makeError(
                    "  with sink '%s' of type %s"
                    % (shortname(sink["id"]), json_dumps(sink["type"])))
            if linkMerge is not None:
                msg += "\n" + SourceLine(sink).makeError("  source has linkMerge method %s" % linkMerge)

        warning_msgs.append(msg)
    for exception in exceptions:
        src = exception.src
        sink = exception.sink
        linkMerge = exception.linkMerge
        msg = SourceLine(src, "type").makeError(
            "Source '%s' of type %s is incompatible"
            % (shortname(src["id"]), json_dumps(src["type"]))) + "\n" + \
            SourceLine(sink, "type").makeError(
                "  with sink '%s' of type %s"
                % (shortname(sink["id"]), json_dumps(sink["type"])))
        if linkMerge is not None:
            msg += "\n" + SourceLine(sink).makeError("  source has linkMerge method %s" % linkMerge)
        exception_msgs.append(msg)

    for sink in step_inputs:
        if ('null' != sink["type"] and 'null' not in sink["type"]
                and "source" not in sink and "default" not in sink and "valueFrom" not in sink):
            msg = SourceLine(sink).makeError(
                "Required parameter '%s' does not have source, default, or valueFrom expression"
                % shortname(sink["id"]))
            exception_msgs.append(msg)

    all_warning_msg = strip_dup_lineno("\n".join(warning_msgs))
    all_exception_msg = strip_dup_lineno("\n".join(exception_msgs))

    if warnings:
        _logger.warning("Workflow checker warning:\n%s", all_warning_msg)
    if exceptions:
        raise validate.ValidationException(all_exception_msg)


SrcSink = namedtuple("SrcSink", ["src", "sink", "linkMerge"])

def check_all_types(src_dict, sinks, sourceField):
    # type: (Dict[Text, Any], List[Dict[Text, Any]], Text) -> Dict[Text, List[SrcSink]]
    """
    Given a list of sinks, check if their types match with the types of their sources.

    sourceField is either "soure" or "outputSource"
    """
    validation = {"warning": [], "exception": []}  # type: Dict[Text, List[SrcSink]]
    for sink in sinks:
        if sourceField in sink:
            valueFrom = sink.get("valueFrom")
            if isinstance(sink[sourceField], MutableSequence):
                srcs_of_sink = [src_dict[parm_id] for parm_id in sink[sourceField]]
                linkMerge = sink.get("linkMerge", ("merge_nested"
                                                   if len(sink[sourceField]) > 1 else None))
            else:
                parm_id = sink[sourceField]
                srcs_of_sink = [src_dict[parm_id]]
                linkMerge = None
            for src in srcs_of_sink:
                check_result = check_types(src, sink, linkMerge, valueFrom)
                if check_result == "warning":
                    validation["warning"].append(SrcSink(src, sink, linkMerge))
                elif check_result == "exception":
                    validation["exception"].append(SrcSink(src, sink, linkMerge))
    return validation



import os

from six.moves import urllib

import pytest
import json

from cwltool.context import LoadingContext
from cwltool.load_tool import load_tool
from cwltool.workflow import default_make_tool
from cwltool.subgraph import get_subgraph
from schema_salad.utils import convert_to_dict
from cwltool.resolver import Path, resolve_local
from .util import get_data, working_directory
from six import string_types
from .test_fetch import norm

def test_get_subgraph():
    loadingContext = LoadingContext({"construct_tool_object": default_make_tool})
    wf = norm(Path(get_data("tests/subgraph/count-lines1-wf.cwl")).as_uri())
    loadingContext.do_update = False
    tool = load_tool(wf, loadingContext)

    sg = norm(Path(get_data("tests/subgraph")).as_uri())

    def clean(val):
        if isinstance(val, string_types):
            if val.startswith(sg):
                return val[len(sg)+1:]
        if isinstance(val, dict):
            return {k: clean(v) for k,v in val.items()}
        if isinstance(val, list):
            return [clean(v) for v in val]
        return val

    for a in ("file1", "file2", "file3", "count_output",
              "output3", "output4", "output5",
              "step1", "step2", "step3", "step4", "step5"):
        extracted = get_subgraph([wf+"#"+a], tool)
        with open(get_data("tests/subgraph/extract_"+a+".json")) as f:
            assert json.load(f) == clean(convert_to_dict(extracted))



"""Helper functions for docker."""
from __future__ import absolute_import, print_function

from typing import List, Optional, Tuple, cast  # pylint: disable=unused-import

from typing_extensions import Text  # pylint: disable=unused-import
# move to a regular typing import when Python 3.3-3.6 is no longer supported

from .utils import subprocess


def docker_vm_id():  # type: () -> Tuple[Optional[int], Optional[int]]
    """
    Return the User ID and Group ID of the default docker user inside the VM.

    When a host is using boot2docker or docker-machine to run docker with
    boot2docker.iso (As on Mac OS X), the UID that mounts the shared filesystem
    inside the VirtualBox VM is likely different than the user's UID on the host.
    :return: A tuple containing numeric User ID and Group ID of the docker account inside
    the boot2docker VM
    """
    if boot2docker_running():
        return boot2docker_id()
    if docker_machine_running():
        return docker_machine_id()
    return (None, None)


def check_output_and_strip(cmd):  # type: (List[Text]) -> Optional[Text]
    """
    Pass a command list to subprocess.check_output.

    Returning None if an expected exception is raised
    :param cmd: The command to execute
    :return: Stripped string output of the command, or None if error
    """
    try:
        result = subprocess.check_output(cmd, stderr=subprocess.STDOUT, universal_newlines=True)
        return cast(str, result).strip()
    except (OSError, subprocess.CalledProcessError, TypeError, AttributeError):
        # OSError is raised if command doesn't exist
        # CalledProcessError is raised if command returns nonzero
        # AttributeError is raised if result cannot be strip()ped
        return None


def docker_machine_name():  # type: () -> Optional[Text]
    """
    Get the machine name of the active docker-machine machine.

    :return: Name of the active machine or None if error
    """
    return check_output_and_strip(['docker-machine', 'active'])


def cmd_output_matches(check_cmd, expected_status):
    # type: (List[Text], Text) -> bool
    """
    Run a command and compares output to expected.

    :param check_cmd: Command list to execute
    :param expected_status: Expected output, e.g. "Running" or "poweroff"
    :return: Boolean value, indicating whether or not command result matched
    """
    return check_output_and_strip(check_cmd) == expected_status


def boot2docker_running():  # type: () -> bool
    """
    Check if boot2docker CLI reports that boot2docker vm is running.

    :return: True if vm is running, False otherwise
    """
    return cmd_output_matches(['boot2docker', 'status'], 'running')


def docker_machine_running():  # type: () -> bool
    """
    Ask docker-machine for the active machine and checks if its VM is running.

    :return: True if vm is running, False otherwise
    """
    machine_name = docker_machine_name()
    if not machine_name:
        return False
    return cmd_output_matches(['docker-machine', 'status', machine_name], 'Running')


def cmd_output_to_int(cmd):  # type: (List[Text]) -> Optional[int]
    """
    Run the provided command and returns the integer value of the result.

    :param cmd: The command to run
    :return: Integer value of result, or None if an error occurred
    """
    result = check_output_and_strip(cmd)  # may return None
    if result is not None:
        try:
            return int(result)
        except ValueError:
            # ValueError is raised if int conversion fails
            return None
    return None


def boot2docker_id():  # type: () -> Tuple[Optional[int], Optional[int]]
    """
    Get the UID and GID of the docker user inside a running boot2docker vm.

    :return: Tuple (UID, GID), or (None, None) if error (e.g. boot2docker not present or stopped)
    """
    uid = cmd_output_to_int(['boot2docker', 'ssh', 'id', '-u'])
    gid = cmd_output_to_int(['boot2docker', 'ssh', 'id', '-g'])
    return (uid, gid)

def docker_machine_id():  # type: () -> Tuple[Optional[int], Optional[int]]
    """
    Ask docker-machine for active machine and gets the UID of the docker user.

    inside the vm
    :return: tuple (UID, GID), or (None, None) if error (e.g. docker-machine not present or stopped)
    """
    machine_name = docker_machine_name()
    if not machine_name:
        return (None, None)
    uid = cmd_output_to_int(['docker-machine', 'ssh', machine_name, "id -u"])
    gid = cmd_output_to_int(['docker-machine', 'ssh', machine_name, "id -g"])
    return (uid, gid)


if __name__ == '__main__':
    print(docker_vm_id())



"""Reference implementation of the CWL standards."""
from __future__ import absolute_import
import warnings
import sys
__author__ = 'pamstutz@veritasgenetics.com'

class CWLToolDeprecationWarning(Warning):
    pass

# Hate this?
# set PYTHONWARNINGS=ignore:DEPRECATION::cwltool.__init__
if sys.version_info < (3, 0):
    warnings.warn("""
DEPRECATION: Python 2.7 will reach the end of its life on January 1st, 2020.
Please upgrade your Python as the Python 2.7 version of cwltool won't be
maintained after that date.
""",  category=CWLToolDeprecationWarning)



from __future__ import absolute_import

import distutils.spawn  # pylint: disable=no-name-in-module,import-error
import functools
import os
import sys
import shutil
import contextlib
import tempfile
import shutil
from typing import Text

from pkg_resources import (Requirement, ResolutionError,  # type: ignore
                           resource_filename)
import pytest

from cwltool.context import LoadingContext, RuntimeContext
from cwltool.factory import Factory
from cwltool.resolver import Path
from cwltool.utils import onWindows, subprocess, windows_default_container_id
from cwltool.singularity import is_version_2_6, is_version_3_or_newer


def get_windows_safe_factory(runtime_context=None,  # type: RuntimeContext
                             loading_context=None,  # type: LoadingContext
                             executor=None          # type: Any
                            ):  # type: (...) -> Factory
    if onWindows():
        if not runtime_context:
            runtime_context = RuntimeContext()
        runtime_context.find_default_container = functools.partial(
            force_default_container, windows_default_container_id)
        runtime_context.use_container = True
        runtime_context.default_container = windows_default_container_id
    return Factory(executor, loading_context, runtime_context)

def force_default_container(default_container_id, _):
    return default_container_id

def get_data(filename):
    # normalizing path depending on OS or else it will cause problem when joining path
    filename = os.path.normpath(filename)
    filepath = None
    try:
        filepath = resource_filename(
            Requirement.parse("cwltool"), filename)
    except ResolutionError:
        pass
    if not filepath or not os.path.isfile(filepath):
        filepath = os.path.join(os.path.dirname(__file__), os.pardir, filename)
    return Text(Path(filepath).resolve())


needs_docker = pytest.mark.skipif(not bool(distutils.spawn.find_executable('docker')),
                                  reason="Requires the docker executable on the "
                                  "system path.")

needs_singularity = pytest.mark.skipif(
    not bool(distutils.spawn.find_executable('singularity')),
    reason="Requires the singularity executable on the system path.")

needs_singularity_2_6 = pytest.mark.skipif(
    not (distutils.spawn.find_executable('singularity') and is_version_2_6()),
    reason="Requires that version 2.6.x of singularity executable version is on the system path.")

needs_singularity_3_or_newer = pytest.mark.skipif(
    not (distutils.spawn.find_executable('singularity') and is_version_3_or_newer),
    reason="Requires that version 2.6.x of singularity executable version is on the system path.")


windows_needs_docker = pytest.mark.skipif(
    onWindows() and not bool(distutils.spawn.find_executable('docker')),
    reason="Running this test on MS Windows requires the docker executable "
    "on the system path.")

def get_main_output(args, env=None):
    process = subprocess.Popen(
        [sys.executable, "-m", "cwltool"] + args,
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env)

    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode(), stderr.decode()

@contextlib.contextmanager
def temp_dir(suffix=""):
    c_dir = tempfile.mkdtemp(suffix, dir=os.curdir)
    try:
        yield c_dir
    finally:
        shutil.rmtree(c_dir, ignore_errors=True)

@contextlib.contextmanager
def working_directory(path):
    """Change working directory and returns to previous on exit."""
    prev_cwd = Path.cwd()
    # before python 3.6 chdir doesn't support paths from pathlib
    os.chdir(str(path))
    try:
        yield
    finally:
        os.chdir(str(prev_cwd))



from cwltool.main import main
from cwltool import load_tool

from .util import get_data, windows_needs_docker


@windows_needs_docker
def test_target():
    """Test --target option successful."""
    test_file = "tests/wf/scatter-wf4.cwl"
    exit_code = main(['--target', 'out', get_data(test_file),
                     '--inp1', 'INP1', '--inp2', 'INP2'])
    assert exit_code == 0


def test_wrong_target():
    """Test --target option when value is wrong."""
    test_file = "tests/wf/scatter-wf4.cwl"
    exit_code = main(['--target', 'dummy_target',
                     get_data(test_file),
                     '--inp1', 'INP1', '--inp2', 'INP2'])
    assert exit_code == 1


@windows_needs_docker
def test_target_packed():
    """Test --target option with packed workflow schema."""
    test_file = "tests/wf/scatter-wf4.json"
    exit_code = main(['--target', 'out',
                     get_data(test_file),
                     '--inp1', 'INP1', '--inp2', 'INP2'])
    assert exit_code == 0



import tempfile
import os
from cwltool.main import main
from cwltool import load_tool
from .util import (get_data, get_windows_safe_factory, windows_needs_docker,
                   needs_docker, temp_dir, needs_singularity, get_main_output)

@windows_needs_docker
def test_newline_in_entry():
    """Files in a InitialWorkingDirectory are created with a newline character."""
    factory = get_windows_safe_factory()
    echo = factory.make(get_data("tests/wf/iwdr-entry.cwl"))
    assert echo(message="hello") == {"out": "CONFIGVAR=hello\n"}


@needs_docker
def test_empty_file_creation():
    """An empty file can be created in InitialWorkingDirectory."""
    err_code, _, _ = get_main_output([get_data("tests/wf/iwdr-empty.cwl")])
    assert err_code == 0


@needs_docker
def test_iwdr_permutations():
    saved_tempdir = tempfile.tempdir
    with temp_dir() as misc:
        tempfile.tempdir = os.path.realpath(misc)
        with temp_dir() as fifth:
    	    with temp_dir() as sixth:
    	        with temp_dir() as seventh:
    	            with temp_dir() as eighth:
    	                with tempfile.NamedTemporaryFile() as first:
    	                    with tempfile.NamedTemporaryFile() as second:
    	                        with tempfile.NamedTemporaryFile() as third:
    	                            with tempfile.NamedTemporaryFile() as fourth:
    	                                with temp_dir() as outdir:
    	                                    assert(main(
    	                                        ['--outdir', outdir,
    	                                         get_data("tests/wf/iwdr_permutations.cwl"),
    	                                         '--first', first.name,
    	                                         '--second', second.name,
    	                                         '--third', third.name,
    	                                         '--fourth', fourth.name,
    	                                         '--fifth', fifth,
    	                                         '--sixth', sixth,
    	                                         '--seventh', seventh,
    	                                         '--eighth', eighth]) == 0)
    tempfile.tempdir = saved_tempdir

@needs_docker
def test_iwdr_permutations_inplace():
    saved_tempdir = tempfile.tempdir
    with temp_dir() as misc:
        tempfile.tempdir = os.path.realpath(misc)
        with temp_dir() as fifth:
            with temp_dir() as sixth:
                with temp_dir() as seventh:
                    with temp_dir() as eighth:
                        with tempfile.NamedTemporaryFile() as first:
                            with tempfile.NamedTemporaryFile() as second:
                                with tempfile.NamedTemporaryFile() as third:
                                    with tempfile.NamedTemporaryFile() as fourth:
                                        with temp_dir() as outdir:
                                            assert(main(
                                                ['--outdir', outdir,
                                                 '--enable-ext',
                                                 '--overrides',
                                                 get_data("tests/wf/iwdr_permutations_inplace.yml"),
                                                 get_data("tests/wf/iwdr_permutations.cwl"),
                                                 '--first', first.name,
                                                 '--second', second.name,
                                                 '--third', third.name,
                                                 '--fourth', fourth.name,
                                                 '--fifth', fifth,
                                                 '--sixth', sixth,
                                                 '--seventh', seventh,
                                                 '--eighth', eighth]) == 0)
    tempfile.tempdir = saved_tempdir

@needs_singularity
def test_iwdr_permutations_singularity():
    with temp_dir() as fifth:
        with temp_dir() as sixth:
            with temp_dir() as seventh:
                with temp_dir() as eighth:
                    with tempfile.NamedTemporaryFile() as first:
                        with tempfile.NamedTemporaryFile() as second:
                            with tempfile.NamedTemporaryFile() as third:
                                with tempfile.NamedTemporaryFile() as fourth:
                                    with temp_dir() as outdir:
                                        assert(main(
                                            ['--outdir', outdir,
                                             '--singularity',
                                             get_data("tests/wf/iwdr_permutations.cwl"),
                                             '--first', first.name,
                                             '--second', second.name,
                                             '--third', third.name,
                                             '--fourth', fourth.name,
                                             '--fifth', fifth,
                                             '--sixth', sixth,
                                             '--seventh', seventh,
                                             '--eighth', eighth]) == 0)

@needs_singularity
def test_iwdr_permutations_singularity_inplace():
    with temp_dir() as fifth:
        with temp_dir() as sixth:
            with temp_dir() as seventh:
                with temp_dir() as eighth:
                    with tempfile.NamedTemporaryFile() as first:
                        with tempfile.NamedTemporaryFile() as second:
                            with tempfile.NamedTemporaryFile() as third:
                                with tempfile.NamedTemporaryFile() as fourth:
                                    with temp_dir() as outdir:
                                        assert(main(
                                            ['--outdir', outdir,
                                             '--singularity',
                                             '--enable-ext',
                                             '--overrides',
                                             get_data("tests/wf/iwdr_permutations_inplace.yml"),
                                             get_data("tests/wf/iwdr_permutations.cwl"),
                                             '--first', first.name,
                                             '--second', second.name,
                                             '--third', third.name,
                                             '--fourth', fourth.name,
                                             '--fifth', fifth,
                                             '--sixth', sixth,
                                             '--seventh', seventh,
                                             '--eighth', eighth]) == 0)



import pytest
import os
from .util import get_data, windows_needs_docker
from cwltool.main import main

@windows_needs_docker
def test_missing_enable_ext():
    # Requires --enable-ext and --enable-dev
    try:
        opt = os.environ.get("CWLTOOL_OPTIONS")

        if "CWLTOOL_OPTIONS" in os.environ:
            del os.environ["CWLTOOL_OPTIONS"]
        assert main([get_data('tests/wf/generator/zing.cwl'),
                     "--zing", "zipper"]) == 1

        assert main(["--enable-ext", "--enable-dev",
                     get_data('tests/wf/generator/zing.cwl'),
                     "--zing", "zipper"]) == 0

        os.environ["CWLTOOL_OPTIONS"] = "--enable-ext --enable-dev"
        assert main([get_data('tests/wf/generator/zing.cwl'),
                     "--zing", "zipper"]) == 0
    finally:
        if opt is not None:
            os.environ["CWLTOOL_OPTIONS"] = opt
        else:
            del os.environ["CWLTOOL_OPTIONS"]



import sys
import json

if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO

from cwltool.main import main

from .util import get_data, needs_docker, temp_dir

@needs_docker
def test_for_910():
    assert main([get_data('tests/wf/910.cwl')]) == 0
    assert main([get_data('tests/wf/910.cwl')]) == 0

@needs_docker
def test_for_conflict_file_names():
    stream = StringIO()

    with temp_dir() as tmp:
        assert main(["--debug", "--outdir", tmp, get_data('tests/wf/conflict.cwl')], stdout=stream) == 0

    out = json.loads(stream.getvalue())
    assert out["b1"]["basename"] == out["b2"]["basename"]
    assert out["b1"]["location"] != out["b2"]["location"]



import json
import logging
import os
import stat
import sys
from io import BytesIO, StringIO
import pytest
from typing_extensions import Text

import schema_salad.validate

import cwltool.checker
import cwltool.expression as expr
import cwltool.factory
import cwltool.pathmapper
import cwltool.process
import cwltool.workflow
from cwltool.context import RuntimeContext
from cwltool.errors import WorkflowException
from cwltool.main import main
from cwltool.utils import onWindows
from cwltool.resolver import Path
from cwltool.process import CWL_IANA
from cwltool.sandboxjs import JavascriptException
from .util import (get_data, get_main_output, get_windows_safe_factory, subprocess,
                   needs_docker, working_directory, needs_singularity, temp_dir, windows_needs_docker)

import six

try:
    reload
except:  # pylint: disable=bare-except
    try:
        from imp import reload
    except:
        from importlib import reload

sys.argv = ['']

expression_match = [
    ("(foo)", True),
    ("(foo.bar)", True),
    ("(foo['bar'])", True),
    ("(foo[\"bar\"])", True),
    ("(foo.bar.baz)", True),
    ("(foo['bar'].baz)", True),
    ("(foo['bar']['baz'])", True),
    ("(foo['b\\'ar']['baz'])", True),
    ("(foo['b ar']['baz'])", True),
    ("(foo_bar)", True),

    ("(foo.[\"bar\"])", False),
    ("(.foo[\"bar\"])", False),
    ("(foo [\"bar\"])", False),
    ("( foo[\"bar\"])", False),
    ("(foo[bar].baz)", False),
    ("(foo['bar\"].baz)", False),
    ("(foo['bar].baz)", False),
    ("{foo}", False),
    ("(foo.bar", False),
    ("foo.bar)", False),
    ("foo.b ar)", False),
    ("foo.b\'ar)", False),
    ("(foo+bar", False),
    ("(foo bar", False)
]


@pytest.mark.parametrize('expression,expected', expression_match)
def test_expression_match(expression, expected):
    match = expr.param_re.match(expression)
    assert (match is not None) == expected


interpolate_input = {
    "foo": {
        "bar": {
            "baz": "zab1"
        },
        "b ar": {
            "baz": 2
        },
        "b'ar": {
            "baz": True
        },
        'b"ar': {
            "baz": None
        }
    },
    "lst": ["A", "B"]
}

interpolate_parameters = [
    ("$(foo)", interpolate_input["foo"]),

    ("$(foo.bar)", interpolate_input["foo"]["bar"]),
    ("$(foo['bar'])", interpolate_input["foo"]["bar"]),
    ("$(foo[\"bar\"])", interpolate_input["foo"]["bar"]),

    ("$(foo.bar.baz)", interpolate_input['foo']['bar']['baz']),
    ("$(foo['bar'].baz)", interpolate_input['foo']['bar']['baz']),
    ("$(foo['bar'][\"baz\"])", interpolate_input['foo']['bar']['baz']),
    ("$(foo.bar['baz'])", interpolate_input['foo']['bar']['baz']),

    ("$(foo['b\\'ar'].baz)", True),
    ("$(foo[\"b'ar\"].baz)", True),
    ("$(foo['b\\\"ar'].baz)", None),

    ("$(lst[0])", "A"),
    ("$(lst[1])", "B"),
    ("$(lst.length)", 2),
    ("$(lst['length'])", 2),

    ("-$(foo.bar)", """-{"baz": "zab1"}"""),
    ("-$(foo['bar'])", """-{"baz": "zab1"}"""),
    ("-$(foo[\"bar\"])", """-{"baz": "zab1"}"""),

    ("-$(foo.bar.baz)", "-zab1"),
    ("-$(foo['bar'].baz)", "-zab1"),
    ("-$(foo['bar'][\"baz\"])", "-zab1"),
    ("-$(foo.bar['baz'])", "-zab1"),

    ("-$(foo['b ar'].baz)", "-2"),
    ("-$(foo['b\\'ar'].baz)", "-true"),
    ("-$(foo[\"b\\'ar\"].baz)", "-true"),
    ("-$(foo['b\\\"ar'].baz)", "-null"),

    ("$(foo.bar) $(foo.bar)", """{"baz": "zab1"} {"baz": "zab1"}"""),
    ("$(foo['bar']) $(foo['bar'])", """{"baz": "zab1"} {"baz": "zab1"}"""),
    ("$(foo[\"bar\"]) $(foo[\"bar\"])", """{"baz": "zab1"} {"baz": "zab1"}"""),

    ("$(foo.bar.baz) $(foo.bar.baz)", "zab1 zab1"),
    ("$(foo['bar'].baz) $(foo['bar'].baz)", "zab1 zab1"),
    ("$(foo['bar'][\"baz\"]) $(foo['bar'][\"baz\"])", "zab1 zab1"),
    ("$(foo.bar['baz']) $(foo.bar['baz'])", "zab1 zab1"),

    ("$(foo['b ar'].baz) $(foo['b ar'].baz)", "2 2"),
    ("$(foo['b\\'ar'].baz) $(foo['b\\'ar'].baz)", "true true"),
    ("$(foo[\"b\\'ar\"].baz) $(foo[\"b\\'ar\"].baz)", "true true"),
    ("$(foo['b\\\"ar'].baz) $(foo['b\\\"ar'].baz)", "null null")
]


@pytest.mark.parametrize('pattern,expected', interpolate_parameters)
def test_expression_interpolate(pattern, expected):
    assert expr.interpolate(pattern, interpolate_input) == expected

interpolate_bad_parameters = [
    ("$(fooz)"),
    ("$(foo.barz)"),
    ("$(foo['barz'])"),
    ("$(foo[\"barz\"])"),

    ("$(foo.bar.bazz)"),
    ("$(foo['bar'].bazz)"),
    ("$(foo['bar'][\"bazz\"])"),
    ("$(foo.bar['bazz'])"),

    ("$(foo['b\\'ar'].bazz)"),
    ("$(foo[\"b'ar\"].bazz)"),
    ("$(foo['b\\\"ar'].bazz)"),

    ("$(lst[O])"),  # not "0" the number, but the letter O
    ("$(lst[2])"),
    ("$(lst.lengthz)"),
    ("$(lst['lengthz'])"),

    ("-$(foo.barz)"),
    ("-$(foo['barz'])"),
    ("-$(foo[\"barz\"])"),

    ("-$(foo.bar.bazz)"),
    ("-$(foo['bar'].bazz)"),
    ("-$(foo['bar'][\"bazz\"])"),
    ("-$(foo.bar['bazz'])"),

    ("-$(foo['b ar'].bazz)"),
    ("-$(foo['b\\'ar'].bazz)"),
    ("-$(foo[\"b\\'ar\"].bazz)"),
    ("-$(foo['b\\\"ar'].bazz)"),
]

@pytest.mark.parametrize('pattern', interpolate_bad_parameters)
def test_expression_interpolate_failures(pattern):
    result = None
    try:
        result = expr.interpolate(pattern, interpolate_input)
    except JavascriptException:
        return
    assert false, 'Should have produced a JavascriptException, got "{}".'.format(result)


@windows_needs_docker
def test_factory():
    factory = get_windows_safe_factory()
    echo = factory.make(get_data("tests/echo.cwl"))

    assert echo(inp="foo") == {"out": "foo\n"}


def test_factory_bad_outputs():
    factory = cwltool.factory.Factory()

    with pytest.raises(schema_salad.validate.ValidationException):
        factory.make(get_data("tests/echo_broken_outputs.cwl"))


def test_factory_default_args():
    factory = cwltool.factory.Factory()

    assert factory.runtime_context.use_container is True
    assert factory.runtime_context.on_error == "stop"


def test_factory_redefined_args():
    runtime_context = RuntimeContext()
    runtime_context.use_container = False
    runtime_context.on_error = "continue"
    factory = cwltool.factory.Factory(runtime_context=runtime_context)

    assert factory.runtime_context.use_container is False
    assert factory.runtime_context.on_error == "continue"


def test_factory_partial_scatter():
    runtime_context = RuntimeContext()
    runtime_context.on_error = "continue"
    factory = cwltool.factory.Factory(runtime_context=runtime_context)

    with pytest.raises(cwltool.factory.WorkflowStatus) as err_info:
        factory.make(get_data("tests/wf/scatterfail.cwl"))()

    err = err_info.value
    assert err.out["out"][0]["checksum"] == 'sha1$e5fa44f2b31c1fb553b6021e7360d07d5d91ff5e'
    assert err.out["out"][1] is None
    assert err.out["out"][2]["checksum"] == 'sha1$a3db5c13ff90a36963278c6a39e4ee3c22e2a436'


def test_factory_partial_output():
    runtime_context = RuntimeContext()
    runtime_context.on_error = "continue"
    factory = cwltool.factory.Factory(runtime_context=runtime_context)

    with pytest.raises(cwltool.factory.WorkflowStatus) as err_info:
        factory.make(get_data("tests/wf/wffail.cwl"))()

    err = err_info.value
    assert err.out["out1"]["checksum"] == 'sha1$e5fa44f2b31c1fb553b6021e7360d07d5d91ff5e'
    assert err.out["out2"] is None


def test_scandeps():
    obj = {
        "id": "file:///example/foo.cwl",
        "steps": [
            {
                "id": "file:///example/foo.cwl#step1",
                "inputs": [{
                    "id": "file:///example/foo.cwl#input1",
                    "default": {
                        "class": "File",
                        "location": "file:///example/data.txt"
                    }
                }],
                "run": {
                    "id": "file:///example/bar.cwl",
                    "inputs": [{
                        "id": "file:///example/bar.cwl#input2",
                        "default": {
                            "class": "Directory",
                            "location": "file:///example/data2",
                            "listing": [{
                                "class": "File",
                                "location": "file:///example/data3.txt",
                                "secondaryFiles": [{
                                    "class": "File",
                                    "location": "file:///example/data5.txt"
                                }]
                            }]
                        },
                    }, {
                        "id": "file:///example/bar.cwl#input3",
                        "default": {
                            "class": "Directory",
                            "listing": [{
                                "class": "File",
                                "location": "file:///example/data4.txt"
                            }]
                        }
                    }, {
                        "id": "file:///example/bar.cwl#input4",
                        "default": {
                            "class": "File",
                            "contents": "file literal"
                        }
                    }]
                }
            }
        ]
    }

    def loadref(base, p):
        if isinstance(p, dict):
            return p
        raise Exception("test case can't load things")

    scanned_deps = cwltool.process.scandeps(
        obj["id"], obj,
        {"$import", "run"},
        {"$include", "$schemas", "location"},
        loadref)

    scanned_deps.sort(key=lambda k: k["basename"])

    expected_deps = [
        {"basename": "bar.cwl",
         "nameroot": "bar",
         "class": "File",
         "format": CWL_IANA,
         "nameext": ".cwl",
         "location": "file:///example/bar.cwl"},
        {"basename": "data.txt",
         "nameroot": "data",
         "class": "File",
         "nameext": ".txt",
         "location": "file:///example/data.txt"},
        {"basename": "data2",
         "class": "Directory",
         "location": "file:///example/data2",
         "listing": [
             {"basename": "data3.txt",
              "nameroot": "data3",
              "class": "File",
              "nameext": ".txt",
              "location": "file:///example/data3.txt",
              "secondaryFiles": [
                  {"class": "File",
                   "basename": "data5.txt",
                   "location": "file:///example/data5.txt",
                   "nameext": ".txt",
                   "nameroot": "data5"
                   }]
              }]
         }, {
            "basename": "data4.txt",
            "nameroot": "data4",
            "class": "File",
            "nameext": ".txt",
            "location": "file:///example/data4.txt"
        }]

    assert scanned_deps == expected_deps

    scanned_deps = cwltool.process.scandeps(
        obj["id"], obj,
        set(("run"), ),
        set(), loadref)

    scanned_deps.sort(key=lambda k: k["basename"])

    expected_deps = [{
        "basename": "bar.cwl",
        "nameroot": "bar",
        "format": CWL_IANA,
        "class": "File",
        "nameext": ".cwl",
        "location": "file:///example/bar.cwl"
    }]

    assert scanned_deps == expected_deps

def test_trick_scandeps():
    if sys.version_info[0] < 3:
        stream = BytesIO()
    else:
        stream = StringIO()

    main(["--print-deps", "--debug", get_data("tests/wf/trick_defaults.cwl")], stdout=stream)
    assert json.loads(stream.getvalue())["secondaryFiles"][0]["location"][:2] != "_:"


def test_input_deps():
    if sys.version_info[0] < 3:
        stream = BytesIO()
    else:
        stream = StringIO()

    main(["--print-input-deps", get_data("tests/wf/count-lines1-wf.cwl"),
          get_data("tests/wf/wc-job.json")], stdout=stream)

    expected = {"class": "File",
                "location": "wc-job.json",
                "format": CWL_IANA,
                "secondaryFiles": [{"class": "File",
                                    "location": "whale.txt",
                                    "basename": "whale.txt",
                                    "nameroot": "whale",
                                    "nameext": ".txt"}]}
    assert json.loads(stream.getvalue()) == expected


def test_input_deps_cmdline_opts():
    if sys.version_info[0] < 3:
        stream = BytesIO()
    else:
        stream = StringIO()

    main(["--print-input-deps",
          get_data("tests/wf/count-lines1-wf.cwl"),
          "--file1", get_data("tests/wf/whale.txt")], stdout=stream)
    expected = {"class": "File",
                "location": "",
                "format": CWL_IANA,
                "secondaryFiles": [{"class": "File",
                                    "location": "whale.txt",
                                    "basename": "whale.txt",
                                    "nameroot": "whale",
                                    "nameext": ".txt"}]}
    assert json.loads(stream.getvalue()) == expected


def test_input_deps_cmdline_opts_relative_deps_cwd():
    if sys.version_info[0] < 3:
        stream = BytesIO()
    else:
        stream = StringIO()

    data_path = get_data("tests/wf/whale.txt")
    main(["--print-input-deps", "--relative-deps", "cwd",
          get_data("tests/wf/count-lines1-wf.cwl"),
          "--file1", data_path], stdout=stream)

    goal = {"class": "File",
            "location": "",
            "format": CWL_IANA,
            "secondaryFiles": [{"class": "File",
                                "location": str(
                                    Path(os.path.relpath(
                                        data_path, os.path.curdir))),
                                "basename": "whale.txt",
                                "nameroot": "whale",
                                "nameext": ".txt"}]}
    assert json.loads(stream.getvalue()) == goal


def test_dedupe():
    not_deduped = [
        {"class": "File",
         "location": "file:///example/a"},
        {"class": "File",
         "location": "file:///example/a"},
        {"class": "File",
         "location": "file:///example/d"},
        {"class": "Directory",
         "location": "file:///example/c",
         "listing": [
             {"class": "File",
              "location": "file:///example/d"}
         ]}
    ]

    expected = [
        {"class": "File",
         "location": "file:///example/a"},
        {"class": "Directory",
         "location": "file:///example/c",
         "listing": [
             {"class": "File",
              "location": "file:///example/d"}
         ]}
    ]

    assert cwltool.pathmapper.dedup(not_deduped) == expected


record = {
    'fields': [
        {'type': {'items': 'string', 'type': 'array'},
         'name': u'file:///home/chapmanb/drive/work/cwl/test_bcbio_cwl/run_info-cwl-workflow/wf-variantcall.cwl#vc_rec/vc_rec/description'
         },
        {'type': {'items': 'File', 'type': 'array'},
         'name': u'file:///home/chapmanb/drive/work/cwl/test_bcbio_cwl/run_info-cwl-workflow/wf-variantcall.cwl#vc_rec/vc_rec/vrn_file'
         }],
    'type': 'record',
    'name': u'file:///home/chapmanb/drive/work/cwl/test_bcbio_cwl/run_info-cwl-workflow/wf-variantcall.cwl#vc_rec/vc_rec'
}

source_to_sink = [
    ('0',
     {'items': ['string', 'null'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     True
     ),
    ('1',
     {'items': ['string'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     True
     ),
    ('2',
     {'items': ['string', 'null'], 'type': 'array'},
     {'items': ['string'], 'type': 'array'},
     True
     ),
    ('3',
     {'items': ['string'], 'type': 'array'},
     {'items': ['int'], 'type': 'array'},
     False
     ),
    ('record 0',
     record, record,
     True
     ),
    ('record 1',
     record, {'items': 'string', 'type': 'array'},
     False
     )
]


@pytest.mark.parametrize('name, source, sink, expected', source_to_sink)
def test_compare_types(name, source, sink, expected):
    assert cwltool.workflow.can_assign_src_to_sink(source, sink) == expected, name


source_to_sink_strict = [
    ('0',
     ['string', 'null'], ['string', 'null'],
     True
     ),
    ('1',
     ['string'], ['string', 'null'],
     True
     ),
    ('2',
     ['string', 'int'], ['string', 'null'],
     False
     ),
    ('3',
     {'items': ['string'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     True
     ),
    ('4',
     {'items': ['string', 'int'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     False
     )
]


@pytest.mark.parametrize('name, source, sink, expected', source_to_sink_strict)
def test_compare_types_strict(name, source, sink, expected):
    assert cwltool.workflow.can_assign_src_to_sink(source, sink, strict=True) == expected, name


typechecks = [
    (['string', 'int'], ['string', 'int', 'null'],
     None, None,
     "pass"
     ),
    (['string', 'int'], ['string', 'null'],
     None, None,
     "warning"
     ),
    (['File', 'int'], ['string', 'null'],
     None, None,
     "exception"
     ),
    ({'items': ['string', 'int'], 'type': 'array'},
     {'items': ['string', 'int', 'null'], 'type': 'array'},
     None, None,
     "pass"
     ),
    ({'items': ['string', 'int'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     None, None,
     "warning"
     ),
    ({'items': ['File', 'int'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     None, None,
     "exception"
     ),
    # check linkMerge when sinktype is not an array
    (['string', 'int'], ['string', 'int', 'null'],
     "merge_nested", None,
     "exception"
     ),
    # check linkMerge: merge_nested
    (['string', 'int'],
     {'items': ['string', 'int', 'null'], 'type': 'array'},
     "merge_nested", None,
     "pass"
     ),
    (['string', 'int'],
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_nested", None,
     "warning"
     ),
    (['File', 'int'],
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_nested", None,
     "exception"
     ),
    # check linkMerge: merge_nested and sinktype is "Any"
    (['string', 'int'], "Any",
     "merge_nested", None,
     "pass"
     ),
    # check linkMerge: merge_flattened
    (['string', 'int'],
     {'items': ['string', 'int', 'null'], 'type': 'array'},
     "merge_flattened", None,
     "pass"
     ),
    (['string', 'int'],
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_flattened", None,
     "warning"
     ),
    (['File', 'int'],
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_flattened", None,
     "exception"
     ),
    ({'items': ['string', 'int'], 'type': 'array'},
     {'items': ['string', 'int', 'null'], 'type': 'array'},
     "merge_flattened", None,
     "pass"
     ),
    ({'items': ['string', 'int'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_flattened", None,
     "warning"
     ),
    ({'items': ['File', 'int'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_flattened", None,
     "exception"),
    # check linkMerge: merge_flattened and sinktype is "Any"
    (['string', 'int'], "Any",
     "merge_flattened", None,
     "pass"
     ),
    ({'items': ['string', 'int'], 'type': 'array'}, "Any",
     "merge_flattened", None,
     "pass"
     ),
    # check linkMerge: merge_flattened when srctype is a list
    ([{'items': 'string', 'type': 'array'}],
     {'items': 'string', 'type': 'array'},
     "merge_flattened", None,
     "pass"
     ),
    # check valueFrom
    ({'items': ['File', 'int'], 'type': 'array'},
     {'items': ['string', 'null'], 'type': 'array'},
     "merge_flattened", "special value",
     "pass"
     )
]


@pytest.mark.parametrize('src_type,sink_type,link_merge,value_from,expected_type', typechecks)
def test_typechecking(src_type, sink_type, link_merge, value_from, expected_type):
    assert cwltool.checker.check_types(
        src_type, sink_type, linkMerge=link_merge, valueFrom=value_from
    ) == expected_type


def test_lifting():
    # check that lifting the types of the process outputs to the workflow step
    # fails if the step 'out' doesn't match.
    factory = cwltool.factory.Factory()
    with pytest.raises(schema_salad.validate.ValidationException):
        echo = factory.make(get_data("tests/test_bad_outputs_wf.cwl"))
        assert echo(inp="foo") == {"out": "foo\n"}


def test_malformed_outputs():
    # check that tool validation fails if one of the outputs is not a valid CWL type
    factory = cwltool.factory.Factory()
    with pytest.raises(schema_salad.validate.ValidationException):
        factory.make(get_data("tests/wf/malformed_outputs.cwl"))()


def test_separate_without_prefix():
    # check that setting 'separate = false' on an inputBinding without prefix fails the workflow
    factory = cwltool.factory.Factory()
    with pytest.raises(WorkflowException):
        factory.make(get_data("tests/wf/separate_without_prefix.cwl"))()


def test_static_checker():
    # check that the static checker raises exception when a source type
    # mismatches its sink type.
    factory = cwltool.factory.Factory()

    with pytest.raises(schema_salad.validate.ValidationException):
        factory.make(get_data("tests/checker_wf/broken-wf.cwl"))

    with pytest.raises(schema_salad.validate.ValidationException):
        factory.make(get_data("tests/checker_wf/broken-wf2.cwl"))

    with pytest.raises(schema_salad.validate.ValidationException):
        factory.make(get_data("tests/checker_wf/broken-wf3.cwl"))


def test_var_spool_cwl_checker1():
    """Confirm that references to /var/spool/cwl are caught."""
    stream = StringIO()
    streamhandler = logging.StreamHandler(stream)
    _logger = logging.getLogger("cwltool")
    _logger.addHandler(streamhandler)

    factory = cwltool.factory.Factory()
    try:
        factory.make(get_data("tests/non_portable.cwl"))
        assert "non_portable.cwl:18:4: Non-portable reference to /var/spool/cwl detected" in stream.getvalue()
    finally:
        _logger.removeHandler(streamhandler)


def test_var_spool_cwl_checker2():
    """Confirm that references to /var/spool/cwl are caught."""
    stream = StringIO()
    streamhandler = logging.StreamHandler(stream)
    _logger = logging.getLogger("cwltool")
    _logger.addHandler(streamhandler)

    factory = cwltool.factory.Factory()
    try:
        factory.make(get_data("tests/non_portable2.cwl"))
        assert "non_portable2.cwl:19:4: Non-portable reference to /var/spool/cwl detected" in stream.getvalue()
    finally:
        _logger.removeHandler(streamhandler)


def test_var_spool_cwl_checker3():
    """Confirm that references to /var/spool/cwl are caught."""
    stream = StringIO()
    streamhandler = logging.StreamHandler(stream)
    _logger = logging.getLogger("cwltool")
    _logger.addHandler(streamhandler)

    factory = cwltool.factory.Factory()
    try:
        factory.make(get_data("tests/portable.cwl"))
        assert "Non-portable reference to /var/spool/cwl detected" not in stream.getvalue()
    finally:
        _logger.removeHandler(streamhandler)


def test_print_dot():
    assert main(["--print-dot", get_data('tests/wf/revsort.cwl')]) == 0

test_factors = [(""), ("--parallel"), ("--debug"), ("--parallel --debug")]

@pytest.mark.parametrize("factor", test_factors)
def test_js_console_cmd_line_tool(factor):
    for test_file in ("js_output.cwl", "js_output_workflow.cwl"):
        commands = factor.split()
        commands.extend(["--js-console", "--no-container", get_data("tests/wf/" + test_file)])
        error_code, _, stderr = get_main_output(commands)

        assert "[log] Log message" in stderr
        assert "[err] Error message" in stderr

        assert error_code == 0, stderr

@pytest.mark.parametrize("factor", test_factors)
def test_no_js_console(factor):
    for test_file in ("js_output.cwl", "js_output_workflow.cwl"):
        commands = factor.split()
        commands.extend(["--no-container", get_data("tests/wf/" + test_file)])
        _, _, stderr = get_main_output(commands)

        assert "[log] Log message" not in stderr
        assert "[err] Error message" not in stderr


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_cid_file_dir(tmpdir, factor):
    test_file = "cache_test_workflow.cwl"
    cwd = tmpdir.chdir()
    commands = factor.split()
    commands.extend(["--cidfile-dir", str(tmpdir), get_data("tests/wf/" + test_file)])
    error_code, stdout, stderr = get_main_output(commands)
    assert "completed success" in stderr
    assert error_code == 0
    cidfiles_count = sum(1 for _ in tmpdir.visit(fil="*"))
    assert cidfiles_count == 2
    cwd.chdir()
    tmpdir.remove(ignore_errors=True)


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_cid_file_dir_arg_is_file_instead_of_dir(tmpdir, factor):
    test_file = "cache_test_workflow.cwl"
    bad_cidfile_dir = Text(tmpdir.ensure("cidfile-dir-actually-a-file"))
    commands = factor.split()
    commands.extend(["--cidfile-dir", bad_cidfile_dir,
         get_data("tests/wf/" + test_file)])
    error_code, _, stderr = get_main_output(commands)
    assert "is not a directory, please check it first" in stderr, stderr
    assert error_code == 2 or error_code == 1, stderr
    tmpdir.remove(ignore_errors=True)


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_cid_file_non_existing_dir(tmpdir, factor):
    test_file = "cache_test_workflow.cwl"
    bad_cidfile_dir = Text(tmpdir.join("cidfile-dir-badpath"))
    commands = factor.split()
    commands.extend(['--record-container-id',"--cidfile-dir", bad_cidfile_dir,
         get_data("tests/wf/" + test_file)])
    error_code, _, stderr = get_main_output(commands)
    assert "directory doesn't exist, please create it first" in stderr, stderr
    assert error_code == 2 or error_code == 1, stderr
    tmpdir.remove(ignore_errors=True)


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_cid_file_w_prefix(tmpdir, factor):
    test_file = "cache_test_workflow.cwl"
    cwd = tmpdir.chdir()
    try:
        commands = factor.split()
        commands.extend(['--record-container-id', '--cidfile-prefix=pytestcid',
             get_data("tests/wf/" + test_file)])
        error_code, stdout, stderr = get_main_output(commands)
    finally:
        listing = tmpdir.listdir()
        cwd.chdir()
        cidfiles_count = sum(1 for _ in tmpdir.visit(fil="pytestcid*"))
        tmpdir.remove(ignore_errors=True)
    assert "completed success" in stderr
    assert error_code == 0
    assert cidfiles_count == 2, '{}/n{}'.format(listing, stderr)


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_secondary_files_v1_1(factor):
    test_file = "secondary-files.cwl"
    test_job_file = "secondary-files-job.yml"
    try:
        old_umask = os.umask(stat.S_IWOTH)  # test run with umask 002
        commands = factor.split()
        commands.extend(["--enable-dev",
            get_data(os.path.join("tests", test_file)),
            get_data(os.path.join("tests", test_job_file))])
        error_code, _, stderr = get_main_output(commands)
    finally:
        assert stat.S_IMODE(os.stat('lsout').st_mode) == 436  # 664 in octal, '-rw-rw-r--'
        os.umask(old_umask)  # revert back to original umask
    assert "completed success" in stderr
    assert error_code == 0

@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_secondary_files_v1_0(factor):
    test_file = "secondary-files-string-v1.cwl"
    test_job_file = "secondary-files-job.yml"
    try:
        old_umask = os.umask(stat.S_IWOTH)  # test run with umask 002
        commands = factor.split()
        commands.extend([
                get_data(os.path.join("tests", test_file)),
                get_data(os.path.join("tests", test_job_file))
            ])
        error_code, _, stderr = get_main_output(commands)
    finally:
        assert stat.S_IMODE(os.stat('lsout').st_mode) == 436  # 664 in octal, '-rw-rw-r--'
        os.umask(old_umask)  # revert back to original umask
    assert "completed success" in stderr
    assert error_code == 0


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_wf_without_container(tmpdir, factor):
    test_file = "hello-workflow.cwl"
    with temp_dir("cwltool_cache") as cache_dir:
        commands = factor.split()
        commands.extend(["--cachedir", cache_dir, "--outdir", str(tmpdir),
             get_data("tests/wf/" + test_file),
             "--usermessage",
             "hello"])
        error_code, _, stderr = get_main_output(commands)

    assert "completed success" in stderr
    assert error_code == 0


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_issue_740_fixed(factor):
    test_file = "cache_test_workflow.cwl"
    with temp_dir("cwltool_cache") as cache_dir:
        commands = factor.split()
        commands.extend(["--cachedir", cache_dir, get_data("tests/wf/" + test_file)])
        error_code, _, stderr = get_main_output(commands)

        assert "completed success" in stderr
        assert error_code == 0

        commands = factor.split()
        commands.extend(["--cachedir", cache_dir, get_data("tests/wf/" + test_file)])
        error_code, _, stderr = get_main_output(commands)

        assert "Output of job will be cached in" not in stderr
        assert error_code == 0, stderr


@needs_docker
def test_compute_checksum():
    runtime_context = RuntimeContext()
    runtime_context.compute_checksum = True
    runtime_context.use_container = onWindows()
    factory = cwltool.factory.Factory(runtime_context=runtime_context)
    echo = factory.make(get_data("tests/wf/cat-tool.cwl"))
    output = echo(
        file1={"class": "File",
               "location": get_data("tests/wf/whale.txt")},
        reverse=False)
    assert output['output']["checksum"] == "sha1$327fc7aedf4f6b69a42a7c8b808dc5a7aff61376"


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_no_compute_chcksum(tmpdir, factor):
    test_file = "tests/wf/wc-tool.cwl"
    job_file = "tests/wf/wc-job.json"
    commands = factor.split()
    commands.extend(["--no-compute-checksum", "--outdir", str(tmpdir),
         get_data(test_file), get_data(job_file)])
    error_code, stdout, stderr = get_main_output(commands)
    assert "completed success" in stderr
    assert error_code == 0
    assert "checksum" not in stdout


@pytest.mark.parametrize("factor", test_factors)
def test_bad_userspace_runtime(factor):
    test_file = "tests/wf/wc-tool.cwl"
    job_file = "tests/wf/wc-job.json"
    commands = factor.split()
    commands.extend([
        "--user-space-docker-cmd=quaquioN", "--default-container=debian",
        get_data(test_file), get_data(job_file)])
    error_code, stdout, stderr = get_main_output(commands)
    assert "or quaquioN is missing or broken" in stderr, stderr
    assert error_code == 1

@windows_needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_bad_basecommand(factor):
    test_file = "tests/wf/missing-tool.cwl"
    commands = factor.split()
    commands.extend([get_data(test_file)])
    error_code, stdout, stderr = get_main_output(commands)
    assert "'neenooGo' not found" in stderr, stderr
    assert error_code == 1


@needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_bad_basecommand_docker(factor):
    test_file = "tests/wf/missing-tool.cwl"
    commands = factor.split()
    commands.extend(
        ["--debug", "--default-container", "debian", get_data(test_file)])
    error_code, stdout, stderr = get_main_output(commands)
    assert "permanentFail" in stderr, stderr
    assert error_code == 1

@pytest.mark.parametrize("factor", test_factors)
def test_v1_0_position_expression(factor):
    test_file = "tests/echo-position-expr.cwl"
    test_job = "tests/echo-position-expr-job.yml"
    commands = factor.split()
    commands.extend(
        ['--debug', get_data(test_file), get_data(test_job)])
    error_code, stdout, stderr = get_main_output(commands)
    assert "is not int" in stderr, stderr
    assert error_code == 1


@windows_needs_docker
@pytest.mark.parametrize("factor", test_factors)
def test_optional_numeric_output_0(factor):
    test_file = "tests/wf/optional-numerical-output-0.cwl"
    commands = factor.split()
    commands.extend([get_data(test_file)])
    error_code, stdout, stderr = get_main_output(commands)

    assert "completed success" in stderr
    assert error_code == 0
    assert json.loads(stdout)['out'] == 0

@pytest.mark.parametrize("factor", test_factors)
@windows_needs_docker
def test_env_filtering(factor):
    test_file = "tests/env.cwl"
    commands = factor.split()
    commands.extend([get_data(test_file)])
    error_code, stdout, stderr = get_main_output(commands)

    process = subprocess.Popen(["sh", "-c", r"""getTrueShellExeName() {
  local trueExe nextTarget 2>/dev/null
  trueExe=$(ps -o comm= $$) || return 1
  [ "${trueExe#-}" = "$trueExe" ] || trueExe=${trueExe#-}
  [ "${trueExe#/}" != "$trueExe" ] || trueExe=$([ -n "$ZSH_VERSION" ] && which -p "$trueExe" || which "$trueExe")
  while nextTarget=$(readlink "$trueExe"); do trueExe=$nextTarget; done
  printf '%s\n' "$(basename "$trueExe")"
} ; getTrueShellExeName"""], stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=None)
    sh_name, sh_name_err = process.communicate()
    sh_name = sh_name.decode('utf-8').strip()

    assert "completed success" in stderr, (error_code, stdout, stderr)
    assert error_code == 0, (error_code, stdout, stderr)
    if onWindows():
        target = 5
    elif sh_name == "dash":
        target = 4
    else:  # bash adds "SHLVL" and "_" environment variables
        target = 6
    result = json.loads(stdout)['env_count']
    details = ''
    if result != target:
        _, details, _ = get_main_output(["--quiet", get_data("tests/env2.cwl")])
        print(sh_name)
        print(sh_name_err)
        print(details)
    assert result == target, (error_code, sh_name, sh_name_err, details, stdout, stderr)

@windows_needs_docker
def test_v1_0_arg_empty_prefix_separate_false():
    test_file = "tests/arg-empty-prefix-separate-false.cwl"
    error_code, stdout, stderr = get_main_output(
        ['--debug', get_data(test_file), "--echo"])
    assert "completed success" in stderr
    assert error_code == 0



import pytest
import sys
import os
import subprocess
from .util import get_data, get_main_output
import tempfile
import shutil
from psutil.tests import TRAVIS

LINUX = sys.platform in ('linux', 'linux2')


#@pytest.mark.skipif(not LINUX, reason="LINUX only")
@pytest.mark.skip("Udocker install is broken, see https://github.com/indigo-dc/udocker/issues/221")
class TestUdocker:
    udocker_path = None

    @classmethod
    def setup_class(cls):
        test_cwd = os.getcwd()
        test_environ = os.environ.copy()
        cls.docker_install_dir = tempfile.mkdtemp()
        os.chdir(cls.docker_install_dir)

        url="https://download.ncg.ingrid.pt/webdav/udocker/udocker-1.1.3.tar.gz"
        install_cmds = [
            ["curl", url, "-o", "./udocker-tarball.tgz"],
            ["tar", "xzvf", "udocker-tarball.tgz", "udocker"],
            ["bash", "-c", "UDOCKER_TARBALL={}/udocker-tarball.tgz ./udocker install".format(cls.docker_install_dir)]]

        os.environ['UDOCKER_DIR'] = os.path.join(cls.docker_install_dir, ".udocker")
        os.environ['HOME'] = cls.docker_install_dir

        results = []
        for _ in range(3):
              results = [subprocess.call(cmds) for cmds in install_cmds]
              if sum(results) == 0:
                  break
              subprocess.call(["rm", "./udocker"])

        assert sum(results) == 0

        cls.udocker_path = os.path.join(cls.docker_install_dir, 'udocker')
        os.chdir(test_cwd)
        os.environ = test_environ
        print('Udocker install dir: ' + cls.docker_install_dir)

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.docker_install_dir)

    def test_udocker_usage_should_not_write_cid_file(self, tmpdir):
        cwd = tmpdir.chdir()

        test_file = "tests/wf/wc-tool.cwl"
        job_file = "tests/wf/wc-job.json"
        error_code, stdout, stderr = get_main_output(
            ["--debug", "--default-container", "debian", "--user-space-docker-cmd=" + self.udocker_path,
             get_data(test_file), get_data(job_file)])
        cwd.chdir()
        cidfiles_count = sum(1 for _ in tmpdir.visit(fil="*.cid"))

        tmpdir.remove(ignore_errors=True)

        assert "completed success" in stderr, stderr
        assert cidfiles_count == 0

    @pytest.mark.skipif(TRAVIS, reason='Not reliable on single threaded test on travis.')
    def test_udocker_should_display_memory_usage(self, tmpdir):
        cwd = tmpdir.chdir()
        error_code, stdout, stderr = get_main_output(
            ["--enable-ext", "--default-container=debian",  "--user-space-docker-cmd=" + self.udocker_path,
             get_data("tests/wf/timelimit.cwl"), "--sleep_time", "10"])
        cwd.chdir()
        tmpdir.remove(ignore_errors=True)

        assert "completed success" in stderr, stderr
        assert "Max memory" in stderr, stderr



import copy
from .process import Process, shortname
from .load_tool import load_tool
from schema_salad import validate
from .errors import WorkflowException
from .context import RuntimeContext, LoadingContext
from typing import (Any, Callable, Dict, Generator, Iterable, List,
                    Mapping, MutableMapping, MutableSequence,
                    Optional, Tuple, Union, cast)
from .loghandler import _logger
from typing_extensions import Text  # pylint: disable=unused-import

class ProcessGeneratorJob(object):
    def __init__(self, procgenerator):
        # type: (ProcessGenerator) -> None
        self.procgenerator = procgenerator
        self.jobout = None         # type: Optional[Dict[Text, Any]]
        self.processStatus = None  # type: Optional[Text]

    def receive_output(self, jobout, processStatus):
        # type: (Dict[Text, Any], Text) -> None
        self.jobout = jobout
        self.processStatus = processStatus

    def job(self,
            job_order,         # type: Mapping[Text, Any]
            output_callbacks,  # type: Callable[[Any, Any], Any]
            runtimeContext     # type: RuntimeContext
           ):  # type: (...) -> Generator[Any, None, None]
        # FIXME: Declare base type for what Generator yields

        try:
            for tool in self.procgenerator.embedded_tool.job(
                    job_order,
                    self.receive_output,
                    runtimeContext):
                yield tool

            while self.processStatus is None:
                yield None

            if self.processStatus != "success":
                output_callbacks(self.jobout, self.processStatus)
                return

            if self.jobout is None:
                raise WorkflowException("jobout should not be None")

            created_tool, runinputs = self.procgenerator.result(job_order, self.jobout, runtimeContext)

            for tool in created_tool.job(
                    runinputs,
                    output_callbacks,
                    runtimeContext):
                yield tool

        except WorkflowException:
            raise
        except Exception as exc:
            _logger.exception("Unexpected exception")
            raise WorkflowException(Text(exc))


class ProcessGenerator(Process):
    def __init__(self,
                 toolpath_object,      # type: MutableMapping[Text, Any]
                 loadingContext        # type: LoadingContext
    ):  # type: (...) -> None
        super(ProcessGenerator, self).__init__(
            toolpath_object, loadingContext)
        self.loadingContext = loadingContext  # type: LoadingContext
        try:
            if isinstance(toolpath_object["run"], MutableMapping):
                self.embedded_tool = loadingContext.construct_tool_object(
                    toolpath_object["run"], loadingContext)  # type: Process
            else:
                loadingContext.metadata = {}
                self.embedded_tool = load_tool(
                    toolpath_object["run"], loadingContext)
        except validate.ValidationException as vexc:
            if loadingContext.debug:
                _logger.exception("Validation exception")
            raise WorkflowException(
                u"Tool definition %s failed validation:\n%s" %
                (toolpath_object["run"], validate.indent(str(vexc))))

    def job(self,
            job_order,         # type: Mapping[Text, Text]
            output_callbacks,  # type: Callable[[Any, Any], Any]
            runtimeContext     # type: RuntimeContext
           ):  # type: (...) -> Generator[Any, None, None]
        # FIXME: Declare base type for what Generator yields
        return ProcessGeneratorJob(self).job(job_order, output_callbacks, runtimeContext)

    def result(self,
            job_order,      # type: Mapping[Text, Any]
            jobout,         # type: Mapping[Text, Any]
            runtimeContext  # type: RuntimeContext
    ):  # type: (...) -> Tuple[Process, MutableMapping[Text, Any]]
        try:
            loadingContext = self.loadingContext.copy()
            loadingContext.metadata = {}
            embedded_tool = load_tool(
                jobout["runProcess"]["location"], loadingContext)
        except validate.ValidationException as vexc:
            if runtimeContext.debug:
                _logger.exception("Validation exception")
            raise WorkflowException(
                u"Tool definition %s failed validation:\n%s" %
                (jobout["runProcess"], validate.indent(str(vexc))))

        if "runInputs" in jobout:
            runinputs = cast(MutableMapping[Text, Any], jobout["runInputs"])
        else:
            runinputs = cast(MutableMapping[Text, Any], copy.deepcopy(job_order))
            for i in self.embedded_tool.tool["inputs"]:
                if shortname(i["id"]) in runinputs:
                    del runinputs[shortname(i["id"])]
            if "id" in runinputs:
                del runinputs["id"]

        return embedded_tool, runinputs



"""Shared functions and other definitions."""
from __future__ import absolute_import

import collections
import os
import platform
import random
import shutil
import string
import sys
import tempfile
from functools import partial  # pylint: disable=unused-import
from typing import (IO, Any, AnyStr, Callable,  # pylint: disable=unused-import
                    Dict, Iterable, List, MutableMapping, MutableSequence,
                    Optional, Union)

import pkg_resources
from mypy_extensions import TypedDict
from schema_salad.utils import json_dump, json_dumps  # pylint: disable=unused-import
from six.moves import urllib, zip_longest
from typing_extensions import Deque, Text  # pylint: disable=unused-import
# move to a regular typing import when Python 3.3-3.6 is no longer supported

from pathlib2 import Path

# no imports from cwltool allowed
if os.name == 'posix':
    if sys.version_info < (3, 5):
        import subprocess32 as subprocess  # nosec # pylint: disable=unused-import
    else:
        import subprocess  # nosec # pylint: disable=unused-import
else:
    import subprocess  # type: ignore  # nosec

windows_default_container_id = "frolvlad/alpine-bash"

Directory = TypedDict('Directory',
                      {'class': Text, 'listing': List[Dict[Text, Text]],
                       'basename': Text})

DEFAULT_TMP_PREFIX = tempfile.gettempdir() + os.path.sep

processes_to_kill = collections.deque()  # type: Deque[subprocess.Popen]

def versionstring():
    # type: () -> Text
    """Version of CWLtool used to execute the workflow."""
    pkg = pkg_resources.require("cwltool")
    if pkg:
        return u"%s %s" % (sys.argv[0], pkg[0].version)
    return u"%s %s" % (sys.argv[0], "unknown version")

def aslist(l):  # type: (Any) -> MutableSequence[Any]
    """Wrap any non-MutableSequence/list in a list."""
    if isinstance(l, MutableSequence):
        return l
    return [l]

def copytree_with_merge(src, dst):  # type: (Text, Text) -> None
    if not os.path.exists(dst):
        os.makedirs(dst)
        shutil.copystat(src, dst)
    lst = os.listdir(src)
    for item in lst:
        spath = os.path.join(src, item)
        dpath = os.path.join(dst, item)
        if os.path.isdir(spath):
            copytree_with_merge(spath, dpath)
        else:
            shutil.copy2(spath, dpath)

def docker_windows_path_adjust(path):
    # type: (Text) -> Text
    r"""
    Adjust only windows paths for Docker.

    The docker run command treats them as unix paths.

    Example: 'C:\Users\foo to /C/Users/foo (Docker for Windows) or /c/Users/foo
    (Docker toolbox).
    """
    if onWindows():
        split = path.split(':')
        if len(split) == 2:
            if platform.win32_ver()[0] in ('7', '8'):  # type: ignore
                split[0] = split[0].lower()  # Docker toolbox uses lowecase windows Drive letters
            else:
                split[0] = split[0].capitalize()
                # Docker for Windows uses uppercase windows Drive letters
            path = ':'.join(split)
        path = path.replace(':', '').replace('\\', '/')
        return path if path[0] == '/' else '/' + path
    return path


def docker_windows_reverse_path_adjust(path):
    # type: (Text) -> (Text)
    r"""
    Change docker path (only on windows os) appropriately back to Windows path.

    Example:  /C/Users/foo to C:\Users\foo
    """
    if path is not None and onWindows():
        if path[0] == '/':
            path = path[1:]
        else:
            raise ValueError("not a docker path")
        splitpath = path.split('/')
        splitpath[0] = splitpath[0]+':'
        return '\\'.join(splitpath)
    return path


def docker_windows_reverse_fileuri_adjust(fileuri):
    # type: (Text) -> (Text)
    r"""
    Convert fileuri to be MS Windows comptabile, if needed.

    On docker in windows fileuri do not contain : in path
    To convert this file uri to windows compatible add : after drive letter,
    so file:///E/var becomes file:///E:/var
    """
    if fileuri is not None and onWindows():
        if urllib.parse.urlsplit(fileuri).scheme == "file":
            filesplit = fileuri.split("/")
            if filesplit[3][-1] != ':':
                filesplit[3] = filesplit[3]+':'
                return '/'.join(filesplit)
            return fileuri
        raise ValueError("not a file URI")
    return fileuri


def onWindows():
    # type: () -> (bool)
    """Check if we are on Windows OS."""
    return os.name == 'nt'


def convert_pathsep_to_unix(path):  # type: (Text) -> (Text)
    """
    Convert path seperators to unix style.

    On windows os.path.join would use backslash to join path, since we would
    use these paths in Docker we would convert it to use forward slashes: /
    """
    if path is not None and onWindows():
        return path.replace('\\', '/')
    return path

def cmp_like_py2(dict1, dict2):  # type: (Dict[Text, Any], Dict[Text, Any]) -> int
    """
    Compare in the same manner as Python2.

    Comparision function to be used in sorting as python3 doesn't allow sorting
    of different types like str() and int().
    This function re-creates sorting nature in py2 of heterogeneous list of
    `int` and `str`
    """
    # extract lists from both dicts
    first, second = dict1["position"], dict2["position"]
    # iterate through both list till max of their size
    for i, j in zip_longest(first, second):
        if i == j:
            continue
        # in case 1st list is smaller
        # should come first in sorting
        if i is None:
            return -1
        # if 1st list is longer,
        # it should come later in sort
        elif j is None:
            return 1

        # if either of the list contains str element
        # at any index, both should be str before comparing
        if isinstance(i, str) or isinstance(j, str):
            return 1 if str(i) > str(j) else -1
        # int comparison otherwise
        return 1 if i > j else -1
    # if both lists are equal
    return 0


def bytes2str_in_dicts(inp  # type: Union[MutableMapping[Text, Any], MutableSequence[Any], Any]
                      ):
    # type: (...) -> Union[Text, MutableSequence[Any], MutableMapping[Text, Any]]
    """
    Convert any present byte string to unicode string, inplace.

    input is a dict of nested dicts and lists
    """
    # if input is dict, recursively call for each value
    if isinstance(inp, MutableMapping):
        for k in inp:
            inp[k] = bytes2str_in_dicts(inp[k])
        return inp

    # if list, iterate through list and fn call
    # for all its elements
    if isinstance(inp, MutableSequence):
        for idx, value in enumerate(inp):
            inp[idx] = bytes2str_in_dicts(value)
            return inp

    # if value is bytes, return decoded string,
    elif isinstance(inp, bytes):
        return inp.decode('utf-8')

    # simply return elements itself
    return inp


def visit_class(rec, cls, op):
    # type: (Any, Iterable[Any], Union[Callable[..., Any], partial[Any]]) -> None
    """Apply a function to with "class" in cls."""
    if isinstance(rec, MutableMapping):
        if "class" in rec and rec.get("class") in cls:
            op(rec)
        for d in rec:
            visit_class(rec[d], cls, op)
    if isinstance(rec, MutableSequence):
        for d in rec:
            visit_class(d, cls, op)

def visit_field(rec, field, op):
    # type: (Any, Text, Union[Callable[..., Any], partial[Any]]) -> None
    """Apply a function to mapping with 'field'."""
    if isinstance(rec, MutableMapping):
        if field in rec:
            rec[field] = op(rec[field])
        for d in rec:
            visit_field(rec[d], field, op)
    if isinstance(rec, MutableSequence):
        for d in rec:
            visit_field(d, field, op)


def random_outdir():  # type: () -> Text
    """Return the random directory name chosen to use for tool / workflow output."""
    # compute this once and store it as a function attribute - each subsequent call will return the same value
    if not hasattr(random_outdir, 'outdir'):
        random_outdir.outdir = '/' + ''.join([random.choice(string.ascii_letters) for _ in range(6)])  # type: ignore  # nosec
    return random_outdir.outdir  # type: ignore

#
# Simple multi-platform (fcntl/msvrt) file locking wrapper
#
try:
    import fcntl

    def shared_file_lock(fd):  # type: (IO[Any]) -> None
        fcntl.flock(fd.fileno(), fcntl.LOCK_SH)

    def upgrade_lock(fd):  # type: (IO[Any]) -> None
        fcntl.flock(fd.fileno(), fcntl.LOCK_EX)

except ImportError:
    import msvcrt

    def shared_file_lock(fd):  # type: (IO[Any]) -> None
        msvcrt.locking(fd.fileno(), msvcrt.LK_LOCK, 1024)  # type: ignore

    def upgrade_lock(fd):  # type: (IO[Any]) -> None
        pass



import json
from ruamel import yaml
import os
import tempfile
from functools import partial

from six import StringIO
import pytest

import cwltool.pack
import cwltool.workflow
from cwltool import load_tool
from cwltool.load_tool import fetch_document, resolve_and_validate_document
from cwltool.main import main, make_relative, print_pack
from cwltool.pathmapper import adjustDirObjs, adjustFileObjs
from cwltool.resolver import tool_resolver
from cwltool.context import LoadingContext

from .util import get_data, needs_docker


def test_pack():
    loadingContext, workflowobj, uri = fetch_document(get_data("tests/wf/revsort.cwl"))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri)[0]

    with open(get_data("tests/wf/expect_packed.cwl")) as packed_file:
        expect_packed = yaml.safe_load(packed_file)

    packed = cwltool.pack.pack(loadingContext.loader, processobj, uri, loadingContext.metadata)
    adjustFileObjs(packed, partial(make_relative, os.path.abspath(get_data("tests/wf"))))
    adjustDirObjs(packed, partial(make_relative, os.path.abspath(get_data("tests/wf"))))

    assert "$schemas" in packed
    assert len(packed["$schemas"]) == len(expect_packed["$schemas"])
    del packed["$schemas"]
    del expect_packed["$schemas"]

    assert packed == expect_packed


def test_pack_single_tool():
    loadingContext, workflowobj, uri = fetch_document(
        get_data("tests/wf/formattest.cwl"))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri)[0]

    packed = cwltool.pack.pack(loadingContext.loader, processobj, uri, loadingContext.metadata)
    assert "$schemas" in packed

def test_pack_rewrites():
    rewrites = {}

    loadingContext, workflowobj, uri = fetch_document(get_data("tests/wf/default-wf5.cwl"))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri)[0]

    cwltool.pack.pack(loadingContext.loader, processobj, uri, loadingContext.metadata, rewrite_out=rewrites)

    assert len(rewrites) == 6

cwl_missing_version_paths = [
    "tests/wf/hello_single_tool.cwl",
    "tests/wf/hello-workflow.cwl"
]

@pytest.mark.parametrize('cwl_path', cwl_missing_version_paths)
def test_pack_missing_cwlVersion(cwl_path):
    """Ensure the generated pack output is not missing the `cwlVersion` in case of single tool workflow and single step workflow."""
    # Testing single tool workflow
    loadingContext, workflowobj, uri = fetch_document(get_data(cwl_path))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri)[0]

    # generate pack output dict
    packed = json.loads(print_pack(loadingContext.loader, processobj, uri, loadingContext.metadata))

    assert packed["cwlVersion"] == 'v1.0'

def test_pack_idempotence_tool():
    """Ensure that pack produces exactly the same document for an already packed CommandLineTool."""
    _pack_idempotently("tests/wf/hello_single_tool.cwl")

def test_pack_idempotence_workflow():
    """Ensure that pack produces exactly the same document for an already packed workflow."""
    _pack_idempotently("tests/wf/count-lines1-wf.cwl")

def _pack_idempotently(document):
    loadingContext, workflowobj, uri = fetch_document(
        get_data(document))
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri)[0]

    # generate pack output dict
    packed = json.loads(print_pack(loadingContext.loader, processobj, uri, loadingContext.metadata))

    loadingContext, workflowobj, uri2 = fetch_document(packed)
    loadingContext.do_update = False
    loadingContext, uri2 = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri2)[0]
    double_packed = json.loads(print_pack(loadingContext.loader, processobj, uri2, loadingContext.metadata))
    assert packed == double_packed

cwl_to_run = [
    ("tests/wf/count-lines1-wf.cwl",
     "tests/wf/wc-job.json",
     False
     ),
    ("tests/wf/formattest.cwl",
     "tests/wf/formattest-job.json",
     True
     ),
]

@needs_docker
@pytest.mark.parametrize('wf_path,job_path,namespaced', cwl_to_run)
def test_packed_workflow_execution(wf_path, job_path, namespaced, tmpdir):
    loadingContext = LoadingContext()
    loadingContext.resolver = tool_resolver
    loadingContext, workflowobj, uri = fetch_document(
        get_data(wf_path), loadingContext)
    loadingContext.do_update = False
    loadingContext, uri = resolve_and_validate_document(
        loadingContext, workflowobj, uri)
    processobj = loadingContext.loader.resolve_ref(uri)[0]
    packed = json.loads(print_pack(loadingContext.loader, processobj, uri, loadingContext.metadata))

    assert not namespaced or "$namespaces" in packed

    wf_packed_handle, wf_packed_path = tempfile.mkstemp()
    with open(wf_packed_path, 'w') as temp_file:
        json.dump(packed, temp_file)

    normal_output = StringIO()
    packed_output = StringIO()

    normal_params = ['--outdir', str(tmpdir), get_data(wf_path), get_data(job_path)]
    packed_params = ['--outdir', str(tmpdir), '--debug', wf_packed_path, get_data(job_path)]

    assert main(normal_params, stdout=normal_output) == 0
    assert main(packed_params, stdout=packed_output) == 0

    assert json.loads(packed_output.getvalue()) == json.loads(normal_output.getvalue())

    os.close(wf_packed_handle)
    os.remove(wf_packed_path)



from __future__ import absolute_import

import os
import shutil
import tempfile
import sys
import re
from io import StringIO

import pytest

from cwltool.main import main
import cwltool.process


from .util import get_data, needs_docker, temp_dir, windows_needs_docker


@needs_docker
def test_missing_enable_ext():
    # Require that --enable-ext is provided.
    assert main([get_data('tests/wf/listing_deep.cwl'), get_data('tests/listing-job.yml')]) != 0

@needs_docker
def test_listing_deep():
    params = ["--enable-ext", get_data('tests/wf/listing_deep.cwl'),
              get_data('tests/listing-job.yml')]
    assert main(params) == 0

@needs_docker
def test_cwltool_options():
    try:
        opt = os.environ.get("CWLTOOL_OPTIONS")
        os.environ["CWLTOOL_OPTIONS"] = "--enable-ext"
        params = [get_data('tests/wf/listing_deep.cwl'),
                  get_data('tests/listing-job.yml')]
        assert main(params) == 0
    finally:
        if opt is not None:
            os.environ["CWLTOOL_OPTIONS"] = opt
        else:
            del os.environ["CWLTOOL_OPTIONS"]

@needs_docker
def test_listing_shallow():
    # This fails on purpose, because it tries to access listing in a subdirectory
    # the same way that listing_deep does, but it shouldn't be expanded.
    params = ["--enable-ext", get_data('tests/wf/listing_shallow.cwl'),
              get_data('tests/listing-job.yml')]
    assert main(params) != 0

@needs_docker
def test_listing_none():
    # This fails on purpose, because it tries to access listing but it shouldn't be there.
    params = ["--enable-ext", get_data('tests/wf/listing_none.cwl'),
              get_data('tests/listing-job.yml')]
    assert main(params) != 0

@needs_docker
def test_listing_v1_0():
    # Default behavior in 1.0 is deep expansion.
    assert main([get_data('tests/wf/listing_v1_0.cwl'), get_data('tests/listing-job.yml')]) == 0

@pytest.mark.skip(reason="This is not the default behaviour yet")
@needs_docker
def test_listing_v1_1():
    # Default behavior in 1.1 will be no expansion
    assert main([get_data('tests/wf/listing_v1_1.cwl'), get_data('tests/listing-job.yml')]) != 0

@needs_docker
def test_double_overwrite(tmpdir):
    with temp_dir() as tmp:
        tmp_name = os.path.join(tmp, "value")

        before_value, expected_value = "1", "3"

        with open(tmp_name, "w") as f:
            f.write(before_value)

        assert main(["--enable-ext", "--outdir", str(tmpdir),
                     get_data('tests/wf/mut2.cwl'), "-a", tmp_name]) == 0

        with open(tmp_name, "r") as f:
            actual_value = f.read()

        assert actual_value == expected_value

@needs_docker
def test_disable_file_overwrite_without_ext():
    with temp_dir() as tmp:
        with temp_dir() as out:
            tmp_name = os.path.join(tmp, "value")
            out_name = os.path.join(out, "value")

            before_value, expected_value = "1", "2"

            with open(tmp_name, "w") as f:
                f.write(before_value)

            assert main(["--outdir", out, get_data('tests/wf/updateval.cwl'), "-r", tmp_name]) == 0

            with open(tmp_name, "r") as f:
                tmp_value = f.read()
            with open(out_name, "r") as f:
                out_value = f.read()

            assert tmp_value == before_value
            assert out_value == expected_value

@needs_docker
def test_disable_dir_overwrite_without_ext():
    with temp_dir() as tmp:
        with temp_dir() as out:

            assert main(["--outdir", out, get_data('tests/wf/updatedir.cwl'), "-r", tmp]) == 0

            assert not os.listdir(tmp)
            assert os.listdir(out)

@needs_docker
def test_disable_file_creation_in_outdir_with_ext():
    with temp_dir() as tmp:
        with temp_dir() as out:

            tmp_name = os.path.join(tmp, "value")
            out_name = os.path.join(out, "value")

            before_value, expected_value = "1", "2"

            with open(tmp_name, "w") as f:
                f.write(before_value)

            params = ["--enable-ext", "--leave-outputs", "--outdir",
                      out, get_data('tests/wf/updateval_inplace.cwl'), "-r", tmp_name]
            assert main(params) == 0

            with open(tmp_name, "r") as f:
                tmp_value = f.read()

            assert tmp_value == expected_value
            assert not os.path.exists(out_name)

@needs_docker
def test_disable_dir_creation_in_outdir_with_ext():
    with temp_dir() as tmp:
        with temp_dir() as out:
            params = ["--enable-ext", "--leave-outputs", "--outdir",
                      out, get_data('tests/wf/updatedir_inplace.cwl'), "-r", tmp]
            assert main(params) == 0

            assert os.listdir(tmp)
            assert not os.listdir(out)

@needs_docker
def test_write_write_conflict():
    with temp_dir('tmp') as tmp:
        tmp_name = os.path.join(tmp, "value")

        before_value, expected_value = "1", "2"

        with open(tmp_name, "w") as f:
            f.write(before_value)

        assert main(["--enable-ext", get_data('tests/wf/mut.cwl'), "-a", tmp_name]) != 0

        with open(tmp_name, "r") as f:
            tmp_value = f.read()

        assert tmp_value == expected_value

@pytest.mark.skip(reason="This test is non-deterministic")
def test_read_write_conflict():
    with temp_dir('tmp') as tmp:
        tmp_name = os.path.join(tmp, "value")

        with open(tmp_name, "w") as f:
            f.write("1")

        assert main(["--enable-ext", get_data('tests/wf/mut3.cwl'), "-a", tmp_name]) != 0

@needs_docker
def test_require_prefix_networkaccess():
    assert main(["--enable-ext", get_data('tests/wf/networkaccess.cwl')]) == 0
    assert main([get_data('tests/wf/networkaccess.cwl')]) != 0
    assert main(["--enable-ext", get_data('tests/wf/networkaccess-fail.cwl')]) != 0

@needs_docker
def test_require_prefix_workreuse(tmpdir):
    assert main(["--enable-ext", '--outdir', str(tmpdir), get_data('tests/wf/workreuse.cwl')]) == 0
    assert main([get_data('tests/wf/workreuse.cwl')]) != 0
    assert main(["--enable-ext", get_data('tests/wf/workreuse-fail.cwl')]) != 0

@windows_needs_docker
def test_require_prefix_timelimit():
    assert main(["--enable-ext", get_data('tests/wf/timelimit.cwl')]) == 0
    assert main([get_data('tests/wf/timelimit.cwl')]) != 0
    assert main(["--enable-ext", get_data('tests/wf/timelimit-fail.cwl')]) != 0

def test_warn_large_inputs():
    was = cwltool.process.FILE_COUNT_WARNING
    try:
        stream = StringIO()

        cwltool.process.FILE_COUNT_WARNING = 3
        main([get_data('tests/wf/listing_v1_0.cwl'), get_data('tests/listing2-job.yml')],
             stderr=stream)

        assert "Recursive directory listing has resulted in a large number of File" in re.sub("\n  *", " ", stream.getvalue())
    finally:
        cwltool.process.FILE_COUNT_WARNING = was



from __future__ import absolute_import

from typing import cast, IO, Any, Dict, MutableMapping

from rdflib import Graph
from schema_salad.jsonld_context import makerdf
from schema_salad.ref_resolver import ContextType
from six.moves import urllib
from typing_extensions import Text  # pylint: disable=unused-import
# move to a regular typing import when Python 3.3-3.6 is no longer supported

from .process import Process


def gather(tool, ctx):  # type: (Process, ContextType) -> Graph
    g = Graph()

    def visitor(t):  # type: (MutableMapping[Text, Any]) -> None
        makerdf(t["id"], t, ctx, graph=g)

    tool.visit(visitor)
    return g


def printrdf(wflow, ctx, style):  # type: (Process, ContextType, str) -> Text
    """Serialize the CWL document into a string, ready for printing."""
    rdf = gather(wflow, ctx).serialize(format=style, encoding='utf-8')
    if not rdf:
        return u""
    return cast(Text, rdf.decode('utf-8'))


def lastpart(uri):  # type: (Any) -> Text
    uri2 = Text(uri)
    if "/" in uri2:
        return uri2[uri2.rindex("/") + 1:]
    return uri2


def dot_with_parameters(g, stdout):  # type: (Graph, IO[Any]) -> None
    qres = g.query(
        """SELECT ?step ?run ?runtype
           WHERE {
              ?step cwl:run ?run .
              ?run rdf:type ?runtype .
           }""")

    for step, run, _ in qres:
        stdout.write(u'"%s" [label="%s"]\n' % (lastpart(step), "%s (%s)" % (lastpart(step), lastpart(run))))

    qres = g.query(
        """SELECT ?step ?inp ?source
           WHERE {
              ?wf Workflow:steps ?step .
              ?step cwl:inputs ?inp .
              ?inp cwl:source ?source .
           }""")

    for step, inp, source in qres:
        stdout.write(u'"%s" [shape=box]\n' % (lastpart(inp)))
        stdout.write(u'"%s" -> "%s" [label="%s"]\n' % (lastpart(source), lastpart(inp), ""))
        stdout.write(u'"%s" -> "%s" [label="%s"]\n' % (lastpart(inp), lastpart(step), ""))

    qres = g.query(
        """SELECT ?step ?out
           WHERE {
              ?wf Workflow:steps ?step .
              ?step cwl:outputs ?out .
           }""")

    for step, out in qres:
        stdout.write(u'"%s" [shape=box]\n' % (lastpart(out)))
        stdout.write(u'"%s" -> "%s" [label="%s"]\n' % (lastpart(step), lastpart(out), ""))

    qres = g.query(
        """SELECT ?out ?source
           WHERE {
              ?wf cwl:outputs ?out .
              ?out cwl:source ?source .
           }""")

    for out, source in qres:
        stdout.write(u'"%s" [shape=octagon]\n' % (lastpart(out)))
        stdout.write(u'"%s" -> "%s" [label="%s"]\n' % (lastpart(source), lastpart(out), ""))

    qres = g.query(
        """SELECT ?inp
           WHERE {
              ?wf rdf:type cwl:Workflow .
              ?wf cwl:inputs ?inp .
           }""")

    for (inp,) in qres:
        stdout.write(u'"%s" [shape=octagon]\n' % (lastpart(inp)))


def dot_without_parameters(g, stdout):  # type: (Graph, IO[Any]) -> None
    dotname = {}  # type: Dict[Text,Text]
    clusternode = {}

    stdout.write("compound=true\n")

    subworkflows = set()
    qres = g.query(
        """SELECT ?run
           WHERE {
              ?wf rdf:type cwl:Workflow .
              ?wf Workflow:steps ?step .
              ?step cwl:run ?run .
              ?run rdf:type cwl:Workflow .
           } ORDER BY ?wf""")
    for (run,) in qres:
        subworkflows.add(run)

    qres = g.query(
        """SELECT ?wf ?step ?run ?runtype
           WHERE {
              ?wf rdf:type cwl:Workflow .
              ?wf Workflow:steps ?step .
              ?step cwl:run ?run .
              ?run rdf:type ?runtype .
           } ORDER BY ?wf""")

    currentwf = None
    for wf, step, run, runtype in qres:
        if step not in dotname:
            dotname[step] = lastpart(step)

        if wf != currentwf:
            if currentwf is not None:
                stdout.write("}\n")
            if wf in subworkflows:
                if wf not in dotname:
                    dotname[wf] = "cluster_" + lastpart(wf)
                stdout.write(u'subgraph "%s" { label="%s"\n' % (dotname[wf], lastpart(wf)))
                currentwf = wf
                clusternode[wf] = step
            else:
                currentwf = None

        if Text(runtype) != "https://w3id.org/cwl/cwl#Workflow":
            stdout.write(u'"%s" [label="%s"]\n' % (dotname[step], urllib.parse.urldefrag(Text(step))[1]))

    if currentwf is not None:
        stdout.write("}\n")

    qres = g.query(
        """SELECT DISTINCT ?src ?sink ?srcrun ?sinkrun
           WHERE {
              ?wf1 Workflow:steps ?src .
              ?wf2 Workflow:steps ?sink .
              ?src cwl:out ?out .
              ?inp cwl:source ?out .
              ?sink cwl:in ?inp .
              ?src cwl:run ?srcrun .
              ?sink cwl:run ?sinkrun .
           }""")

    for src, sink, srcrun, sinkrun in qres:
        attr = u""
        if srcrun in clusternode:
            attr += u'ltail="%s"' % dotname[srcrun]
            src = clusternode[srcrun]
        if sinkrun in clusternode:
            attr += u' lhead="%s"' % dotname[sinkrun]
            sink = clusternode[sinkrun]
        stdout.write(u'"%s" -> "%s" [%s]\n' % (dotname[src], dotname[sink], attr))


def printdot(wf, ctx, stdout, include_parameters=False):
    # type: (Process, ContextType, Any, bool) -> None
    g = gather(wf, ctx)

    stdout.write("digraph {")

    # g.namespace_manager.qname(predicate)

    if include_parameters:
        dot_with_parameters(g, stdout)
    else:
        dot_without_parameters(g, stdout)

    stdout.write("}")



from cwltool.main import main

from .util import get_data


def test_missing_cwl_version():
    """No cwlVersion in the workflow."""
    assert main([get_data('tests/wf/missing_cwlVersion.cwl')]) == 1

def test_incorrect_cwl_version():
    """Using cwlVersion: v0.1 in the workflow."""
    assert main([get_data('tests/wf/wrong_cwlVersion.cwl')]) == 1
